from abc import ABC, abstractmethod
import importlib
import json
import os
from pathlib import Path
import sys
from typing import Any, Dict, Generator, List, Union

import arrow
import frontmatter
import jinja2
import markdown


def chunks(lst: List, n: int) -> Generator[List, None, None]:
    """Yield successive n-sized chunks from lst

    Thank you SO: https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
    """

    for i in range(0, len(lst), n):
        yield lst[i : i + n]


class SitePage:
    def __init__(
        self, source: "PageFile", data: Dict, collections: Dict[str, List["PageFile"]]
    ) -> None:
        self.source = source
        self.data = data
        self.collections = collections
        self.template_data = {"data": self.data, "collections": self.collections}

    def render(self, jinja2_env: jinja2.Environment) -> str:
        html = self.source.get_html(jinja2_env, page=self, **self.template_data)
        if "layout" in self.data:
            template = jinja2_env.get_template(self.data["layout"])
            return template.render(
                content=html,
                page=self,
                data=self.data,
                collections=self.collections,
            )
        else:
            return html

    def get_path(self) -> Path:
        if "permalink" in self.data:

            jinja2_env = jinja2.Environment()

            permalink_template = jinja2_env.from_string(self.data["permalink"])
            permalink = permalink_template.render(data=self.data)
            if permalink.endswith("/"):
                # Filename is not specified, so append 'index.html'
                return Path(permalink) / "index.html"
            else:
                # Assume the last part of the path is the filename
                return Path(permalink)
        else:
            path = self.source.path
            if len(path.parts) == 1 and path.stem == "index":
                # Anything just named "index" in the top level dir becomes the
                # index page
                return Path("index.html")
            else:
                return path.parent / path.stem / "index.html"

    def get_permalink(self) -> str:
        permalink = str(self.get_path())

        if not permalink.startswith("/"):
            permalink = "/" + permalink

        if permalink.endswith("index.html"):
            permalink = permalink[: -1 * len("index.html")]
        return permalink


class PaginationSitePage(SitePage):
    def __init__(
        self,
        source: "PageFile",
        data: Dict,
        collections: Dict[str, List["PageFile"]],
        index: int,
        items: List,
    ) -> None:
        super().__init__(source, data, collections)
        self.index = index
        self.items = items
        self.template_data = {
            **self.template_data,
            "items": self.items,
            "index": self.index,
        }

    def get_path(self) -> Path:
        if self.index == 0:
            return super().get_path()
        else:
            if "permalink" in self.data:
                jinja2_env = jinja2.Environment()
                permalink_template = jinja2_env.from_string(self.data["permalink"])
                permalink = permalink_template.render(
                    data=self.data, index=self.index, items=self.items
                )
                if permalink.endswith("/"):
                    # Filename is not specified, so append 'index.html'
                    return Path(permalink) / "index.html"
                else:
                    # Assume the last part of the path is the filename
                    return Path(permalink)
            else:
                path = self.source.path
                return path.parent / path.stem / str(self.index) / "index.html"


class InvalidFileExtensionException(Exception):
    pass


class InvalidTagsException(Exception):
    pass


class MissingPaginationSourceException(Exception):
    pass


class SourceFile(ABC):
    suffixes: set

    def __init__(self, path: Path) -> None:
        if path.suffix not in self.suffixes:
            raise InvalidFileExtensionException(path.suffix)
        self.path = path
        self.date = arrow.get(os.path.getmtime(self.path))

    def __str__(self):
        return str(self.path)


class NoPermalinkException(Exception):
    pass


class PageFile(SourceFile):
    def __init__(self, path: Path, data: Dict) -> None:
        super().__init__(path)

        with open(self.path) as infile:
            metadata, self.content = frontmatter.parse(infile.read())

        self.data = {**data, **metadata}

        if "tags" in self.data:
            if isinstance(self.data["tags"], str):
                self.tags = set()
                self.tags.add(self.data["tags"])
            elif isinstance(self.data["tags"], List):
                self.tags = set(self.data["tags"])
            else:
                raise InvalidTagsException(
                    f"Invalid tags for file at {self.path}. Expected <str> or <list>, "
                    f"got {type(self.data['tags'])}"
                )
        else:
            self.tags = set()

        if "date" in self.data:
            self.date = arrow.get(self.data["date"])

    def get_pages(self, collections: Dict[str, List["PageFile"]]) -> List[SitePage]:
        if "pagination" in self.data:

            pagination_source = self.data["pagination"]["data"]

            if pagination_source in self.data:
                pagination_data = self.data[pagination_source]
            elif pagination_source in collections:
                pagination_data = collections[pagination_source]
            else:
                raise MissingPaginationSourceException(pagination_source)

            pages: List[SitePage] = []
            for index, items in enumerate(
                chunks(pagination_data, self.data["pagination"]["size"])
            ):
                pages.append(
                    PaginationSitePage(self, self.data, collections, index, items)
                )
            return pages
        else:
            return [SitePage(self, self.data, collections)]

    def get_permalink(self):
        if "pagination" in self.data:
            raise NoPermalinkException(
                "Cannot get a permalink from a PageFile with pagination"
            )

        return SitePage(self, self.data, []).get_permalink()

    def get_html(self, jinja2_env, **kwargs):
        return self.content


class MarkdownFile(PageFile):
    suffixes = {".md"}

    def get_html(self, _, **kwargs):
        return markdown.markdown(self.content, extensions=["codehilite", "fenced_code"])


class Jinja2File(PageFile):
    def get_html(self, jinja2_env, **kwargs):
        template = jinja2_env.from_string(self.content)
        return template.render(**kwargs)

    suffixes = {".html", ".j2"}


class DataFile(SourceFile, ABC):
    @abstractmethod
    def get_data(self):  # pragma: no cover
        return


class JSONFile(DataFile):
    suffixes = {".json"}

    def get_data(self) -> Union[List, Dict]:
        with open(self.path) as infile:
            return json.load(infile)


class PythonFile(DataFile):
    suffixes = {".py"}

    def get_data(self) -> Any:
        parent_path = str(self.path.parent)
        if self.path.parent.name != "":
            sys.path.append(parent_path)
            module = __import__(self.path.stem)
            importlib.reload(module)
            data = module.get_data()
            sys.path.remove(parent_path)

            return data


class DataFileFactory:
    suffix_to_class_map = {".json": JSONFile, ".py": PythonFile}

    def is_valid_file(self, path: Path) -> bool:
        return path.suffix in self.suffix_to_class_map and (
            path.suffix != ".py" or path.parent.name != ""
        )

    def load_source_file(self, path: Path) -> DataFile:
        suffix = path.suffix
        if not self.is_valid_file(path):
            raise InvalidFileExtensionException(
                f"No DataFile type found with suffix {suffix}"
            )
        return self.suffix_to_class_map[suffix](path)


class PageFileFactory:
    suffix_to_class_map = {".html": Jinja2File, ".md": MarkdownFile, ".j2": Jinja2File}

    def is_valid_file(self, path: Path) -> bool:
        return path.suffix in self.suffix_to_class_map

    def load_source_file(self, path: Path, data: Dict) -> PageFile:
        suffix = path.suffix
        if not self.is_valid_file(path):
            raise InvalidFileExtensionException(
                f"No PageFile type found with suffix {suffix}"
            )

        return self.suffix_to_class_map[suffix](path, data)



from os import DirEntry

from watchgod import AllWatcher, DefaultWatcher


class SkipDefaultWatcher(DefaultWatcher):
    def __init__(self, root_path: str) -> None:
        self.ignored_dirs.add("_site")
        super().__init__(root_path)


class SkipIgnoreWatcher(AllWatcher):
    def __init__(self, root_path: str, should_ignore=None) -> None:
        self.should_ignore = should_ignore
        super().__init__(root_path)

    def should_watch_dir(self, entry: DirEntry) -> bool:
        return not self.should_ignore(entry.path)

    def should_watch_file(self, entry: DirEntry) -> bool:
        return not self.should_ignore(entry.path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['skip_ssg']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.1,<4.0.0',
 'Markdown>=3.3.4,<4.0.0',
 'Pygments>=2.9.0,<3.0.0',
 'arrow>=1.1.1,<2.0.0',
 'gitignore-parser>=0.0.8,<0.0.9',
 'python-frontmatter>=1.0.0,<2.0.0',
 'watchgod>=0.7,<0.8']

entry_points = \
{'console_scripts': ['skip = skip_ssg.skip:main']}

setup_kwargs = {
    'name': 'skip-ssg',
    'version': '0.2.0',
    'description': 'A simple python-based static site generator',
    'long_description': "# Skip\n\nSkip is a Python-based data-forward static site generator inspired by [Eleventy](https://11ty.dev)\n\n## Why should I use Skip?\n\n- You like Python and the Jinja2 templating engine\n- You want a simple static site generator that makes no assumptions about how your project should be structured\n- You want to use data files, including dynamically computed data (like reading from a database or API)\n\n## Quickstart\n\n``` bash\npip install skip-ssg\necho '# Hello World' > index.md\nskip\n```\n\nTo see more, check out the [documentation](http://skip.harryeldridge.com)\n",
    'author': 'Harry Eldridge',
    'author_email': 'heldridge@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/heldridge/skip',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import argparse
from collections import defaultdict
import errno
import os
from pathlib import Path
import shutil
from typing import Any, Callable, Dict, List, Mapping, Set

from gitignore_parser import parse_gitignore
import jinja2
import watchgod

import skip_ssg.server as server
from skip_ssg.sources import (
    DataFile,
    DataFileFactory,
    PageFile,
    PageFileFactory,
)
import skip_ssg.watchers as watchers

try:
    import settings
except ImportError:
    print("No settings file found, using defaults")
    USE_SETTINGS = False


def write_page(
    site_dir: Path, permalink: Path, filepath: Path, html: str, quiet: bool = False
) -> None:
    full_path = site_dir / permalink
    if not quiet:
        print("Writing", full_path, "from", filepath)
    os.makedirs(full_path.parent, exist_ok=True)
    with open(full_path, "w+") as outfile:
        outfile.write(html)


def get_data_from_datafiles(data_files: List[DataFile], fail_on_error) -> Dict:
    data = {}
    for data_file in data_files:
        try:
            file_data = data_file.get_data()
        except Exception as e:
            if not fail_on_error:
                print(f'Failed to load data from {data_file.path}: "{e}", skipping...')
                continue
            else:
                raise e

        if not isinstance(file_data, Dict):
            file_data = {data_file.path.stem: file_data}

        data = {**data, **file_data}
    return data


def get_page_files(
    ignores: Set[str],
    should_ignore: Callable[[str], bool],
    path: Path,
    data: Dict,
    fail_on_error: bool,
) -> List[PageFile]:
    page_paths = []
    page_files: List[PageFile] = []
    data_files: List[DataFile] = []
    dirs = []

    pff = PageFileFactory()
    dff = DataFileFactory()
    # Go over all the files to identiy all the data and page sources
    for entry in os.scandir(path):
        if entry.name in ignores or should_ignore(entry.path):
            continue
        if entry.is_dir():
            dirs.append(Path(entry.path))
        elif entry.is_file():
            path = Path(entry.path)
            if pff.is_valid_file(path):
                # Don't load pages yet because we need to process all the data first
                page_paths.append(path)
            elif dff.is_valid_file(path):
                data_files.append(dff.load_source_file(path))

        data = {**data, **get_data_from_datafiles(data_files, fail_on_error)}

    for page_path in page_paths:
        page_files.append(pff.load_source_file(page_path, data))

    # Recurse
    for dir_path in dirs:
        page_files += get_page_files(
            ignores, should_ignore, dir_path, data, fail_on_error
        )

    return page_files


def get_collections(pages: List[PageFile]) -> Mapping[str, List[PageFile]]:
    collections = defaultdict(list)
    for page in pages:
        collections["all"].append(page)
        for tag in page.tags:
            collections[tag].append(page)

    for collection in collections.values():
        collection.sort(key=lambda item: item.date)

    return collections


def false(_: Any) -> bool:
    return False


def build_site(config: Dict, should_ignore: Callable[[str], bool]) -> None:
    print("Building Site")

    data = {}

    site_dir = Path(config["output"])

    ignore_dirs = {
        ".git",
        "data",
        config["output"],
        "templates",
        "__pycache__",
        "venv",
        ".venv",
        "node_modules",
    }

    dff = DataFileFactory()
    data_files = []
    for root, dirs, files in os.walk("data"):
        if "__pycache__" in dirs:
            del dirs[dirs.index("__pycache__")]

        for file in files:
            path = Path(root, file)
            if dff.is_valid_file(path):
                data_files.append(dff.load_source_file(path))

    data = get_data_from_datafiles(data_files, config["fail_on_error"])

    page_files = get_page_files(
        ignore_dirs, should_ignore, Path("."), data, config["fail_on_error"]
    )
    collections = get_collections(page_files)

    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader("templates"))
    for page_file in page_files:
        pages = page_file.get_pages(collections)
        for page in pages:
            html = page.render(jinja_env)
            write_page(site_dir, page.get_path(), page.source.path, html)

    for copy_target in config["copy"]:
        if ":" in copy_target:
            src, dest = copy_target.split(":")
        else:
            src = dest = copy_target

        dest = site_dir / dest
        print(f"Copying {src} to {dest}")

        try:
            shutil.copytree(src, dest, dirs_exist_ok=True)
        except OSError as e:
            if e.errno == errno.ENOTDIR:
                shutil.copy(src, dest)
            else:
                raise e

    print("Build Complete!\n")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-w", "--watch", help="Watch files and reload on changes", action="store_true"
    )
    parser.add_argument(
        "-s", "--serve", help="Serve the site on localhost", action="store_true"
    )
    parser.add_argument("-p", "--port", help="The port to serve on", type=int)
    parser.add_argument(
        "-o",
        "--output",
        help="The directory to output the built site to",
    )
    parser.add_argument(
        "-c",
        "--copy",
        help=(
            "Files or directories to copy into the root of site folder. Use a colon to "
            'specify a new path. For example, "x.css:style/x.css" will copy the file '
            'to "./_site/style/x.css"'
        ),
        nargs="+",
    )
    parser.add_argument(
        "-f",
        "--fail-on-error",
        help=(
            "Terminate with an error when a data file fails to load rather than "
            "skipping it"
        ),
        action="store_true",
    )
    args = parser.parse_args()

    skipignore_path = Path(".skipignore")
    if skipignore_path.exists():
        print("Using .skipignore...")
        should_ignore = parse_gitignore(skipignore_path)
    else:
        print("No .skipignore found, using defaults...")
        should_ignore = false

    # Default config
    config = {"output": "_site", "copy": []}

    if USE_SETTINGS:
        config = {**config, **vars(settings).get("OPTIONS", {})}

    # CLI flags take precedence over settings file
    dict_args = vars(args)
    arg_config_options = ["output", "port", "copy", "fail_on_error"]
    for option in arg_config_options:
        if dict_args[option] is not None:
            config[option] = dict_args[option]

    build_site(config, should_ignore)

    if args.serve:
        server.run(config)

    if args.watch or args.serve:
        print("\nWatching files for changes...")

        if should_ignore is not false:
            for changes in watchgod.watch(
                ".",
                watcher_cls=watchers.SkipIgnoreWatcher,
                watcher_kwargs={"should_ignore": should_ignore},
            ):
                build_site(config, should_ignore)
        else:
            for changes in watchgod.watch(".", watcher_cls=watchers.SkipDefaultWatcher):
                build_site(config, should_ignore)


if __name__ == "__main__":
    main()



import errno
import http.server
import functools
import socketserver
import threading
from typing import Callable, Dict


class QuietHander(http.server.SimpleHTTPRequestHandler):
    def log_message(self, format, *args):
        # Silence all logs
        return


def _start_server_on_port(
    handler: Callable[..., http.server.SimpleHTTPRequestHandler], port: int
):
    with socketserver.TCPServer(("", port), handler) as httpd:
        print(f"Serving at localhost:{port}")
        httpd.serve_forever()


def run_server(config: Dict):
    Handler = functools.partial(QuietHander, directory=config["output"])

    if config.get("port") is None:
        current_port = 8080
        while current_port <= 65535:
            try:
                _start_server_on_port(Handler, current_port)
            except OSError as e:
                # errno 48: Address already in use
                if e.errno == errno.EADDRINUSE:
                    current_port += 1
                    continue
                else:
                    raise e

    else:
        _start_server_on_port(Handler, config["port"])


def run(config: Dict):
    server_thread = threading.Thread(target=run_server, args=(config,), daemon=True)
    server_thread.start()
